#pragma once

#define JP2RAW_REG_PATH _T("Software\\GPSoftware\\Directory Opus\\Config\\User\\Viewers\\JP2Raw")

namespace NJP2RawConfig
{
	class CDropTarget : public IDropTarget
	{
	protected:
		bool	m_bDragEnabled;
		std::vector< HWND > m_vecDropWindows;
		std::vector< std::basic_string<TCHAR> > m_vecFilenames;
		HWND	m_hwnd;
	public:
		CDropTarget();
		virtual ~CDropTarget();
		void SetDragEnabled(bool bEnabled);
		void AddDropWindow(HWND hWnd);
		void Register(HWND hwnd);
		void Revoke();
		std::vector< std::basic_string<TCHAR> > &GetFilenames();
		HWND GetDropWindowFromScreenPoint(POINT &ptScreen);

		// IUnknown
		virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);
		virtual ULONG STDMETHODCALLTYPE AddRef(void);
		virtual ULONG STDMETHODCALLTYPE Release(void);
		// IDropTarget
        virtual HRESULT STDMETHODCALLTYPE DragEnter(IDataObject *pDataObj, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
		virtual HRESULT STDMETHODCALLTYPE DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
		virtual HRESULT STDMETHODCALLTYPE DragLeave(void);
		virtual HRESULT STDMETHODCALLTYPE Drop(IDataObject *pDataObj, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	};

	enum NJR_Ability
	{
		NTTA_RUN	= 0,
	};

	enum NJR_Flags
	{
		NJRF_DISABLE_DCRAW	= 1<<0,
		NJRF_DISABLE_TGA	= 1<<1,
		NJRF_DISABLE_JP2	= 1<<2,
		NJRF_DISABLE_PNM	= 1<<3,
		NJRF_DISABLE_RAS	= 1<<4,
		NJRF_DISABLE_TIF	= 1<<5,
	};

	enum ComboFlagsDisplayedImage
	{
		CF_DI_PREVIEW	= 1<<0,
		CF_DI_FULL		= 1<<1,
	};



	struct NJR_RawSettings
	{
		// Default copy-const/assignment-op must work or be overridden.

		bool	bTryPreview;
		bool	bTryFull;
		int		iRotation;
		bool	bFujiTurn;
		int		iInterpQuality;
		bool	bDocModeNoCol;
		bool	bDocModeRaw;
		bool	bHalfSizeColor;
		bool	bRGGB;
		bool	bUseSecondary;
		bool	bBilateral;
		int		iBilatDomain;
		int		iBilatRange;
		bool	bAutoWhite;
		bool	bCamWhite;
		bool	bSetWhite;
		int		iUserMul1;
		int		iUserMul2;
		int		iUserMul3;
		int		iUserMul4;
		bool	bSetBlackPoint;
		int		iBlackPoint;
		int		iBrightness;
		int		iHighlightMode;
		int		iOutIccType;
		std::string strOutIcc;
		int		iCamIccType;
		std::string strCamIcc;

		NJR_RawSettings()
		: bTryPreview(true)
		, bTryFull(false)
		, iRotation(-1)
		, bFujiTurn(true)
		, iInterpQuality(-1)
		, bDocModeNoCol(false)
		, bDocModeRaw(false)
		, bHalfSizeColor(false)
		, bRGGB(false)
		, bUseSecondary(false)
		, bBilateral(false)
		, iBilatDomain(2000)
		, iBilatRange(4000)
		, bAutoWhite(false)
		, bCamWhite(false)
		, bSetWhite(false)
		, iUserMul1(1000)
		, iUserMul2(1000)
		, iUserMul3(1000)
		, iUserMul4(1000)
		, bSetBlackPoint(false)
		, iBlackPoint(0)
		, iBrightness(1000)
		, iHighlightMode(0)
		, iOutIccType(1)
		, iCamIccType(0)
		{
		}
	};

	struct NJR_RawProfile
	{
		std::string		strDisplayName;
		std::string		strSafeName;
		NJR_RawSettings rsThumbs;
		NJR_RawSettings rsViewers;
	};

	// map key is the "safe" version of the display-name.
	class NJR_ProfileMap : public std::map< std::string, NJR_RawProfile >
	{
	public:
		NJR_RawProfile &GetCameraProfileFromMakeModel(const char *szMake, const char *szModel)
		{
			return GetCameraProfileFromDisplayName(GenerateDisplayName(szMake, szModel).c_str());
		}

		NJR_RawProfile &GetCameraProfileFromDisplayName(const char *szDisplayName)
		{
			return GetCameraProfileFromSafeName(NLeoHelpers::GenerateSafeName(szDisplayName).c_str());
		}

		NJR_RawProfile &GetCameraProfileFromSafeName(const char *szSafeName)
		{
			iterator pPair = find(szSafeName);

			if (pPair == end())
			{
				pPair = find(""); // Find the default profile.
			}

			if (pPair == end())
			{
				throw "ProfileMap has no default";
			}

			return pPair->second;
		}

		static std::string GenerateDisplayName(const char *szMake, const char *szModel)
		{
			std::string strOutDisplayName = szMake;
			if (!strOutDisplayName.empty()) { strOutDisplayName += " / "; }
			strOutDisplayName += szModel;
			return strOutDisplayName;
		}
	};

	struct NJR_ConfigData
	{
		HINSTANCE		hInstance;
		CRITICAL_SECTION *pLcmsCS;
		HWND			hWndNotify;
		DWORD			dwNotifyData;
		DWORD64			dw64OpusVersion;
		HWND			hWndMainDlg;
		HWND			hWndFormats;
		HWND			hWndDCRawThumbs;
		HWND			hWndDCRawViewers;
		HWND			hWndWithTongue;
		bool			bWithTongue;
		NJR_ProfileMap	mapProfiles;
		CDropTarget		dropTarget;
		bool			bProcessingWMScroll;
		std::string		strSelectedProfileSafeName;
		std::string     strActivePage;

		NJR_ConfigData()
		: hInstance(NULL)
		, pLcmsCS(NULL)
		, hWndNotify(NULL)
		, dwNotifyData(0)
		, dw64OpusVersion(0)
		, hWndMainDlg(NULL)
		, hWndFormats(NULL)
		, hWndDCRawThumbs(NULL)
		, hWndDCRawViewers(NULL)
		, hWndWithTongue(NULL)
		, bWithTongue(false)
		, bProcessingWMScroll(false)
		{
		}
	};

	BOOL CALLBACK configDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	BOOL CALLBACK childProxyDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	BOOL CALLBACK tongueDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT CALLBACK editFloatFilterSubProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

	void configInitDialog(BOOL &bResult, bool &bDeletePCD, bool &bUpdateControls, bool &bLoadSettings, bool &bLoadDefaultSettings, bool &bSaveSettings, bool &bEnableDisable, NJR_ConfigData *&pcd, HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	void configNotify(    BOOL &bResult, bool &bDeletePCD, bool &bUpdateControls, bool &bLoadSettings, bool &bLoadDefaultSettings, bool &bSaveSettings, bool &bEnableDisable, NJR_ConfigData * pcd, HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	void configVScroll(   BOOL &bResult, bool &bDeletePCD, bool &bUpdateControls, bool &bLoadSettings, bool &bLoadDefaultSettings, bool &bSaveSettings, bool &bEnableDisable, NJR_ConfigData * pcd, HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	void configCommand(   BOOL &bResult, bool &bDeletePCD, bool &bUpdateControls, bool &bLoadSettings, bool &bLoadDefaultSettings, bool &bSaveSettings, bool &bEnableDisable, NJR_ConfigData * pcd, HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
	void configDropFiles( BOOL &bResult, bool &bDeletePCD, bool &bUpdateControls, bool &bLoadSettings, bool &bLoadDefaultSettings, bool &bSaveSettings, bool &bEnableDisable, NJR_ConfigData * pcd, HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);

	void configLoadSettings( NJR_ConfigData *pcd, bool bLoadDefaultSettings, bool bChangeSelections);
	void configSaveSettings( NJR_ConfigData *pcd);
	void configEnableDisable(NJR_ConfigData *pcd);


	void SetupRawDialog(NJR_ConfigData *pcd, bool bThumbs);
	void SetupRawProfileInDialog(NJR_ConfigData *pcd, NJR_RawProfile &rawProfile, bool bThumbs);

	void StoreRawDialogData(NJR_ConfigData *pcd, bool bThumbs);

	void EnableDisableRawDialog(NJR_ConfigData *pcd, BOOL bRawEnabled, bool bThumbs);

	void SetupComboItem(HWND hWndParent, int iDlgItem, bool bClear, int iValue, int iSelectValue, const char *szTitle);
	void SetupCheckBoxItem(HWND hWndParent, int iDlgItem, bool bCheck);
	void SetupSpinItem(HWND hWndParent, int iDlgItem, int iMax, int iMin, int iValue);
	void SetupEditItem(HWND hWndParent, int iDlgItem, const char *szText);

	void GetComboItem(HWND hWndParent, int iDlgItem, int &iValue);
	void GetCheckBoxItem(HWND hWndParent, int iDlgItem, bool &bValue);
	void GetSpinItem(HWND hWndParent, int iDlgItem, int &iValue);
	void GetEditItem(HWND hWndParent, int iDlgItem, std::string &strText);
	void FilenameToEdit(NJR_ConfigData *pcd, HWND hwndDlg, int iDlgItem, char *szTitle);

	DWORD RegGetFlags(bool bGetDefaultValue);
	void RegGetRawExtensions(bool bGetDefaultValue, std::vector< std::basic_string<TCHAR> > *pvecResult);
	void RegGetRawProfiles(bool bGetDefaultValue, NJR_ProfileMap &mapProfiles);
	void RegGetRawProfile(bool bGetDefaultValue, NJR_RawProfile &rawProfile, const char *szSafeName);
	void RegGetRawSettings(bool bGetDefaultValue, const std::string &strSettingsPath, NJR_RawSettings &rawSettings, bool bThumbs);

	void RegSaveRawProfiles(NJR_ProfileMap &mapProfiles);
	void RegSaveRawSettings(const std::string &settingsPath, const NJR_RawSettings &rs);


	bool CheckOpus(DWORD64 dw64OpusVersion, NJR_Ability ability);
};
