#pragma once

#if 0
#include <time.h>
#include <Shlobj.h>
#pragma message("********************************************************")
#pragma message("********************************************************")
#pragma message("**       Compiled with debug logging turned on.       **")
#pragma message("********************************************************")
#pragma message("********************************************************")
namespace NLeoHelpers {	TCHAR *JoinPaths(const TCHAR *first, const TCHAR *second); };
inline void logLine(const char *szFormat, ...)
{
	va_list args;
	va_start(args, szFormat);

	TCHAR szDesktopPath[MAX_PATH];

	if (SHGetSpecialFolderPath(GetDesktopWindow(), szDesktopPath, CSIDL_DESKTOPDIRECTORY, FALSE))
	{
		char *szLogPath = NLeoHelpers::JoinPaths(szDesktopPath, "jp2raw_debug.txt");

		FILE *pFile = fopen(szLogPath, "a");
		if (NULL != pFile)
		{
			time_t timer;
			time(&timer);
			char szTime[28];

			const struct tm *timmy = localtime(&timer);

			if (NULL != timmy && 0 != strftime(szTime, 28, "[%H:%M:%S] ", timmy))
			{
				fprintf(pFile, "%s", szTime);
			}

			vfprintf(pFile, szFormat, args);
			fclose(pFile);
		}

		delete [] szLogPath;
	}

	va_end(args);
}
#endif

#ifndef MONITOR_DEFAULTTONULL
#define MONITOR_DEFAULTTONULL       0x00000000
#endif
#ifndef MONITOR_DEFAULTTOPRIMARY
#define MONITOR_DEFAULTTOPRIMARY    0x00000001
#endif
#ifndef MONITOR_DEFAULTTONEAREST
#define MONITOR_DEFAULTTONEAREST    0x00000002
#endif

class CxImage;

namespace NLeoHelpers
{
	// Returns false on failure. Use GetLastError().
	// hKeyParent -- hKey of parent under which szKeyPath should be created. Usually a hive like HKEY_CURRENT_USER
	// szKeyPath -- Sub key (can be a path) under hKeyHiveOrParent from which szValueName is read. Can be "" or NULL.
	// szValueName -- Name of value to query. Can be "" or NULL to read the unnamed default value.
	// pdwType -- Pointer to DWORD to receive type of data. NULL if unwanted.
	// plpData -- POINTER TO POINTER to receive data buffer. NULL if data unwanted.
	// pcbData -- Pointer to DWORD to receive size of data buffer. NULL if data unwanted.
	// You do NOT allocate a buffer, this function does it for you, like the frigging API should.
	// If the function succeeds and plpData was not NULL you must delete[] *plpData when finished with it.
	// Like RegQueryValueEx, you may supply NULL plpData and non-NULL pcbData to get the required
	// buffer size (except for HKEY_PERFORMANCE_DATA), but again the function always allocates a buffer for you.
	// Although untested, reading HKEY_PERFORMANCE_DATA should work.
	bool LeetRegQueryValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName,
						DWORD *pdwType, void **plpData, DWORD *pcbData);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegQueryValue() to get you a DWORD without having to worry about buffers, types and so on.
	bool LeetRegQueryDWORDValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, DWORD *pdwRes);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegQueryValue() to get you an int without having to worry about buffers, types and so on.
	bool LeetRegQueryIntValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, int *piRes);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegQueryValue() to get you a bool without having to worry about buffers, types and so on.
	bool LeetRegQueryBoolValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, bool *pBool);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegQueryValue() to get you a TCHAR string without having to worry about buffers, types and so on.
	bool LeetRegQueryStringValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, std::basic_string<TCHAR> *pString);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegQueryValue() to get you a vector of TCHAR strings without having to worry about buffers, types and so on.
	bool LeetRegQueryMultiStringValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, std::vector< std::basic_string<TCHAR> > *pVecStrings);

	// Returns false on failure. Use GetLastError().
	// hKeyParent -- hKey of parent under which szKeyPath should be created. Usually a hive like HKEY_CURRENT_USER
	// szKeyPath -- Sub key (can be a path) under hKeyHiveOrParent in which szValueName is set. Can be "" or NULL.
	// szValueName -- Name of value to set. Can be "" or NULL to set the unnamed, default value.
	// dwType -- Type of data to set. (Same as dwType argument of RegSetValueEx())
	// lpData -- Pointer to data to set. (6/Feb/2001: Can be NULL if you just want to create the key but set no values.)
	// cbData -- Size of data pointed to by lpData, including null terminator(s) (Same as cbData arg of RegSetValueEx())
	bool LeetRegSetValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName,
							DWORD dwType, const void *lpData, DWORD cbData);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegSetValue() to set a DWORD without having to worry about buffers, types and so on.
	bool LeetRegSetDWORDValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, DWORD dwData);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegSetValue() to set an int without having to worry about buffers, types and so on.
	bool LeetRegSetIntValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, int iData);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegSetValue() to set a bool without having to worry about buffers, types and so on.
	bool LeetRegSetBoolValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, bool bData);

	// Returns false on failure. Use GetLastError().
	// Uses LeetRegSetValue() to set a String without having to worry about buffers, types and so on.
	bool LeetRegSetStringValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, const TCHAR *szData);

	// Returns false on failure. Use GetLastError().
	bool LeetRegSetMultiStringValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, const std::vector< std::basic_string<TCHAR> > *pVecStrings);

	// Returns false on failure. Use GetLastError().
	// Returns vector of child key names.
	// If szPrefixFilter is non-NULL then only key names starting with it will be included, and it will be removed from the start of them.
	bool LeetRegEnumKey(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szPrefixFilter, std::vector< std::basic_string<TCHAR> > *pVecStrings);

	// Returns false on failure. Use GetLastError().
	// Will NOT delete a key that has sub-keys. (You can use SHDeleteKey for that, though.)
	bool LeetRegDeleteKey(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szKeyToDelete);

	// strncpy doesn't always null term the destination... sigh.
	// This returns the index of the null terminator.
	int StringCopy(TCHAR *dest, const TCHAR *source, int destSizeChars);

	// _snprintf doesn't always null term the destination... sigh
	// Returns the index of the null terminator.
	int StringFormat(TCHAR *buffer, size_t bufferSizeChars, const TCHAR *format, ...);

	// _vsnprintf doesn't always null term the destination... sigh
	// Returns the index of the null terminator.
	int VStringFormat(TCHAR *buffer, size_t bufferSizeChars, const TCHAR *format, va_list args);

	// delete[] the result.
	// Returns NULL on failure.
	TCHAR *StringAllocAndFormat(const TCHAR *format, ...);

	// delete[] the result.
	// Returns NULL on failure.
	TCHAR *VStringAllocAndFormat(const TCHAR *format, va_list args);

	// You must delete[] the result of this function after use.
	// A path with no slashes or the empty string returns the empty string (you must still delete the result).
	TCHAR *GetParentPath(const TCHAR *path);

	// You must delete[] the result of this function after use.
	TCHAR *JoinPaths(const TCHAR *first, const TCHAR *second);

	HANDLE OpenTempFileNamePreserveExtension(const TCHAR *szTempPath, const TCHAR *szPrefix, const TCHAR *szFilenameIn,
											 std::basic_string< TCHAR > *pstrFilenameOut);

	std::string GenerateSafeName(const char *szInput);

	// Returns NULL on failure (use GetLastError())
	// delete[] the result when finished with it.
	WCHAR *LeetMBtoWC(const char *mbString);

	// Returns NULL on failure (use GetLastError())
	// delete[] the result when finished with it.
	char *LeetWCtoMB(const WCHAR *wcString);

	const TCHAR *GetLastPathPart(const TCHAR *szIn);

	BOOL ScreenToClientRect(HWND hwnd, RECT *pRect);

	//BOOL OffsetDlgItem(HWND hwndDlg, int nIDDlgItem, LONG dx, LONG dy, LONG dw, LONG dh, BOOL bRepaint);
	BOOL EnableDlgItem(HWND hwndDlg, int nIDDlgItem, BOOL bEnable);
	BOOL GetDlgItemText(HWND hwndDlg, int nIDDlgItem, std::basic_string< TCHAR > *pStrResult);
	BOOL GetListBoxItemText(HWND hwndListBox, LRESULT lrItemIndex, std::basic_string< TCHAR > *pStrResult);
	//void SubclassDlgItemToUserData(HWND hwndDlg, int nIDDlgItem, WNDPROC wndProc);
	BOOL GetDlgItemLong(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG *plResult);
	//BOOL SetDlgItemLong(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG lValue);
	//BOOL GetDlgItemLongPtr(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG_PTR *plResult);
	//BOOL SetDlgItemLongPtr(HWND hwndDlg, int nIDDlgItem, int nIndex, LONG_PTR lValue);
	//BOOL ModifyDlgItemStyle(HWND hwndDlg, int nIDDlgItem, bool bStyleEx, LONG lAdd, LONG lRemove);
	//BOOL InvalidateDlgItem(HWND hwndDlg, int nIDDlgItem);


	// Give hWndRelative NULL to centre on the monitor work area with the mouse pointer.
	// On systems before Windows 2000 and Windows 98 which don't support multiple monitors,
	// the window centers on the desktop if hWndRelative is NULL.
	// The window is snapped to the monitor work area it appears on.
	void CenterWindow(HWND hWndToMove, HWND hWndRelative);

	// Regardless of dwFlags, if the API cannot be loaded this function returns NULL.
	HMONITOR MonitorFromPoint(POINT pt, DWORD dwFlags);

	BOOL GetMonitorInfo(HMONITOR hMonitor, LPMONITORINFO lpmi);

	// On systems without multi-monitor support, returns the primary monitor areas.
	// On systems with multi-monitor support, returns the areas for window's monitor.
	// Flags as for MonitorFromWindow.
	// If you don't want a rect pass in NULL.
	bool GetAreasFromWindow(RECT *pRectFull, RECT *pRectWork, HWND hWnd, DWORD dwFlags);

	HBITMAP MakeBitmapWithAlpha(CxImage &xi, bool bMakeAllOpaque);
};

/*

//////////////////////////////////////////////////////////////////
//
//	ExpandShortcut()
//
//	Uses IShellLink to expand a shortcut.
//
//	Return value:
//		the expanded filename, or "" on error or if filename
//		wasn't a shortcut
//
//	Adapted from CShortcut, 1996 by Rob Warner
//	rhwarner@southeast.net
//	http://users.southeast.net/~rhwarner

CString CFileDropListCtrl::ExpandShortcut(CString& csFilename) const
{
	USES_CONVERSION;		// For T2COLE() below
	CString csExpandedFile;

	//
    // Make sure we have a path
	//
	if(csFilename.IsEmpty())
	{
		ASSERT(FALSE);
		return csExpandedFile;
	}

	//
    // Get a pointer to the IShellLink interface
	//
    HRESULT hr;
    IShellLink* pIShellLink;

    hr = ::CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
							IID_IShellLink, (LPVOID*) &pIShellLink);

    if (SUCCEEDED(hr))
    {

		//
        // Get a pointer to the persist file interface
		//
		IPersistFile* pIPersistFile;
        hr = pIShellLink->QueryInterface(IID_IPersistFile, (LPVOID*) &pIPersistFile);

        if (SUCCEEDED(hr))
        {
			//
            // Load the shortcut and resolve the path
			//
            // IPersistFile::Load() expects a UNICODE string
			// so we're using the T2COLE macro for the conversion
			//
			// For more info, check out MFC Technical note TN059
			// (these macros are also supported in ATL and are
			// so much better than the ::MultiByteToWideChar() family)
			//
            hr = pIPersistFile->Load(T2COLE(csFilename), STGM_READ);
			
			if (SUCCEEDED(hr))
			{
				WIN32_FIND_DATA wfd;
				hr = pIShellLink->GetPath(csExpandedFile.GetBuffer(MAX_PATH),
										  MAX_PATH,
										  &wfd,
										  SLGP_UNCPRIORITY);

				csExpandedFile.ReleaseBuffer(-1);
            }
            pIPersistFile->Release();
        }
        pIShellLink->Release();
    }

    return csExpandedFile;
}

*/