#include "stdafx.h"
#include "LeoThemeHelper.h"

// From uxtheme.h

#ifndef THEMESIZE
typedef enum THEMESIZE
{
	TS_MIN,             // minimum size
	TS_TRUE,            // size without stretching
	TS_DRAW,            // size that theme mgr will use to draw part
};
#endif

#ifndef ETDT_DISABLE
#define ETDT_DISABLE        0x00000001
#endif

#ifndef ETDT_ENABLE
#define ETDT_ENABLE         0x00000002
#endif

#ifndef ETDT_USETABTEXTURE
#define ETDT_USETABTEXTURE  0x00000004
#endif

#ifndef ETDT_ENABLETAB
#define ETDT_ENABLETAB      (ETDT_ENABLE  | ETDT_USETABTEXTURE)
#endif

#ifndef TABP_BODY
#define TABP_BODY 10
#endif

CLeoThemeHelper::CLeoThemeHelper()
{
	m_hmodUxThemeDLL = LoadLibrary(_T("uxtheme.dll"));

	if (NULL == m_hmodUxThemeDLL
	||	NULL == (m_fOpenThemeData			 = reinterpret_cast<OPENTHEMEDATA_PTR			>(GetProcAddress(m_hmodUxThemeDLL, _T("OpenThemeData"))))
	||	NULL == (m_fCloseThemeData			 = reinterpret_cast<CLOSETHEMEDATA_PTR			>(GetProcAddress(m_hmodUxThemeDLL, _T("CloseThemeData"))))
	||	NULL == (m_fDrawThemeBackground		 = reinterpret_cast<DRAWTHEMEBACKGROUND_PTR		>(GetProcAddress(m_hmodUxThemeDLL, _T("DrawThemeBackground"))))
	||	NULL == (m_fGetThemePartSize		 = reinterpret_cast<GETTHEMEPARTSIZE_PTR		>(GetProcAddress(m_hmodUxThemeDLL, _T("GetThemePartSize"))))
	||	NULL == (m_fEnableThemeDialogTexture = reinterpret_cast<ENABLETHEMEDIALOGTEXTURE_PTR>(GetProcAddress(m_hmodUxThemeDLL, _T("EnableThemeDialogTexture")))))
	{
		if (NULL != m_hmodUxThemeDLL)
		{
			FreeLibrary(m_hmodUxThemeDLL);
			m_hmodUxThemeDLL = NULL;
		}
		
		m_fOpenThemeData = NULL;
		m_fCloseThemeData = NULL;
		m_fDrawThemeBackground = NULL;
		m_fGetThemePartSize = NULL;
		m_fEnableThemeDialogTexture = NULL;
	}

	m_hbmpDIB = NULL;
}

CLeoThemeHelper::~CLeoThemeHelper()
{
	if (NULL != m_hbmpDIB)
	{
		DeleteObject(m_hbmpDIB);
		m_hbmpDIB = NULL;
	}

	if (NULL != m_hmodUxThemeDLL)
	{
		FreeLibrary(m_hmodUxThemeDLL);
		m_hmodUxThemeDLL = NULL;
	}
}

bool CLeoThemeHelper::EnableTabThemeDialogTexture(HWND hWnd)
{
	bool bResult = false;

	if (NULL != m_hmodUxThemeDLL)
	{
		if (SUCCEEDED(m_fEnableThemeDialogTexture(hWnd, ETDT_ENABLETAB)))
		{
			bResult = true;
		}
	}

	return(bResult);
}

bool CLeoThemeHelper::IsAvailable()
{
	return(NULL != m_hmodUxThemeDLL);
}

bool CLeoThemeHelper::CacheBackgroundTabThemeDialog(HWND hWnd)
{
	bool bResult = false;

	if (NULL != m_hbmpDIB)
	{
		DeleteObject(m_hbmpDIB);
		m_hbmpDIB = NULL;
	}

	if (IsAvailable())
	{
		HANDLE hTheme = m_fOpenThemeData(hWnd, L"Tab");

		if (NULL != hTheme)
		{
			HDC hdcMem = CreateCompatibleDC(NULL);

			if (NULL != hdcMem)
			{
				if (SUCCEEDED(m_fGetThemePartSize(hTheme, hdcMem, TABP_BODY, 0, NULL, TS_TRUE, &m_sizePart)))
				{
					BITMAPINFO bmi;
					ZeroMemory(&bmi, sizeof(bmi));
					bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
					bmi.bmiHeader.biWidth = m_sizePart.cx;
					bmi.bmiHeader.biHeight = m_sizePart.cy * 2;
					bmi.bmiHeader.biPlanes = 1;
					bmi.bmiHeader.biBitCount = 24;
					bmi.bmiHeader.biCompression = BI_RGB;

					void *pDontCare = NULL;

					m_hbmpDIB = CreateDIBSection(hdcMem, &bmi, DIB_RGB_COLORS, &pDontCare, NULL, 0);

					if (NULL != m_hbmpDIB)
					{
						HBITMAP hbmpOld = NULL;

						if (NULL != (hbmpOld = reinterpret_cast<HBITMAP>(SelectObject(hdcMem, m_hbmpDIB))))
						{
							// Get an upside-down copy of the bitmap so that we can draw over every second
							// line of right-way-up bitmaps, so that the stupid thing tiles.

							RECT rectBackgroud;
							rectBackgroud.left	 = 0;
							rectBackgroud.top	 = 0;
							rectBackgroud.right	 = m_sizePart.cx;
							rectBackgroud.bottom = m_sizePart.cy;

							if (SUCCEEDED(m_fDrawThemeBackground(hTheme, hdcMem, TABP_BODY, 0, &rectBackgroud, NULL)))
							{
								StretchBlt(hdcMem, 0, m_sizePart.cy * 2 - 1, m_sizePart.cx, -m_sizePart.cy, hdcMem,
												   0, 0,					 m_sizePart.cx,  m_sizePart.cy, SRCCOPY);

								bResult = true;
							}

							SelectObject(hdcMem, hbmpOld);
						}
					}
				}

				DeleteDC(hdcMem);
			}

			m_fCloseThemeData(hTheme);
		}
	}

	if (!bResult)
	{
		if (NULL != m_hbmpDIB)
		{
			DeleteObject(m_hbmpDIB);
			m_hbmpDIB = NULL;
		}
	}

	return(bResult);
}

void CLeoThemeHelper::EraseExtraBackgroundTabThemeDialog(HWND hWnd, HDC hDC, RECT &rcClient) 
{
	if (NULL != m_hbmpDIB || CacheBackgroundTabThemeDialog(hWnd))
	{
		HBITMAP hbmpOld = NULL;

		HDC hdcMem = CreateCompatibleDC(hDC);

		if (NULL != hdcMem)
		{
			if (NULL != (hbmpOld = reinterpret_cast<HBITMAP>(SelectObject(hdcMem, m_hbmpDIB))))
			{
				for (int y = rcClient.top + m_sizePart.cy; y < rcClient.bottom; y += (m_sizePart.cy * 2))
				{
					for (int x = rcClient.left; x < rcClient.right; x += m_sizePart.cx)
					{
						BitBlt(hDC, x, y, m_sizePart.cx, m_sizePart.cy, hdcMem, 0, m_sizePart.cy, SRCCOPY);
					}
				}

				SelectObject(hdcMem, hbmpOld);
			}

			DeleteDC(hdcMem);
		}
	}
}
