#include "stdafx.h"
#include "Win32IOWrapper.h"

CWin32IOWrapper::CWin32IOWrapper()
: m_bThrowExceptions(false)
{
}

CWin32IOWrapper::~CWin32IOWrapper()
{
}

bool CWin32IOWrapper::GetThrowExceptions()
{
	return m_bThrowExceptions;
}

void CWin32IOWrapper::SetThrowExceptions(bool bThrow)
{
	m_bThrowExceptions = bThrow;
}

int CWin32IOWrapper::getc()
{
	char c;
	if (1 != read(&c, sizeof(char), 1))
	{
		return EOF;
	}
	return static_cast<unsigned char>(c);
}

char *CWin32IOWrapper::gets(char *string, int n)
{
	char *result = string;

	if (0 >= n)
	{
		if (m_bThrowExceptions) { throw "gets given empty buffer"; }
		result = NULL;
	}
	else
	{
		char c;
		char *p = string;

		while (1 < n)
		{
			if (1 != read(&c, sizeof(char), 1))
			{
				result = NULL;
				break;
			}

			if (c == '\r' || c == '\n')
			{
				*p++ = '\n'; n--;
				break;
			}

			*p++ = c; n--;
		}

		*p = '\0';
	}

	return result;
}

int CWin32IOWrapper::getInt()
{
	std::string tempString;
	long seekCorrection = 0;

	while(true)
	{
		bool readError = false;
		char c = '\0';

		try
		{
			if (1 != read(&c, sizeof(char), 1))
			{
				readError = true;
			}
		}
		catch(...)
		{
			readError = true;
		}

		if (readError)
		{
			break;
		}

		if (isdigit(c))
		{
			seekCorrection = 0;
			tempString += c;
		}
		else if (tempString.empty() && (c == '-' || c == '+'))
		{
			seekCorrection -= 1;
			tempString += c;
		}
		else
		{
			seekCorrection -= 1;
			seek(seekCorrection, SEEK_CUR);
			break;
		}
	}

	if (tempString.empty())
	{
		if (m_bThrowExceptions)
		{
			throw "Could not read integer from input";
		}
		else
		{
			return 0;
		}
	}
	else
	{
		return atoi(tempString.c_str());
	}
}

float CWin32IOWrapper::getFloat()
{
	std::string tempString;
	long seekCorrection = 0;

	char lastChar = '\0';
	bool reachedExponent = false;
	bool reachedDecmial = false;

	// [sign] [digits] [.digits] [ {d | D | e | E }[sign]digits]

	while(true)
	{
		bool readError = false;
		char c = '\0';

		try
		{
			if (1 != read(&c, sizeof(char), 1))
			{
				readError = true;
			}
		}
		catch(...)
		{
			readError = true;
		}

		if (readError)
		{
			break;
		}

		if (isdigit(c))
		{
			seekCorrection = 0;
			tempString += c;
			lastChar = c;
		}
		else if (c == '.' && !reachedDecmial && !reachedExponent)
		{
			seekCorrection -= 1;
			tempString += c;
			lastChar = c;
			reachedDecmial = true;
		}
		else if ((c == 'd' || c == 'D' || c == 'e' || c == 'E') && !reachedExponent && isdigit(lastChar))
		{
			seekCorrection -= 1;
			tempString += c;
			lastChar = c;
			reachedExponent = true;
		}
		else if ((c == '-' || c == '+') && (lastChar == '\0' || lastChar == 'd' || lastChar == 'D' || lastChar == 'e' || lastChar == 'E'))
		{
			seekCorrection -= -1;
			tempString += c;
			lastChar = c;
		}
		else
		{
			seekCorrection -= 1;
			seek(seekCorrection, SEEK_CUR);
			break;
		}
	}

	if (tempString.empty())
	{
		if (m_bThrowExceptions)
		{
			throw "Could not read float from input";
		}
		else
		{
			return 0.0f;
		}
	}
	else
	{
		return static_cast<float>(atof(tempString.c_str()));
	}
}

/*

This doesn't work, doesn't compile and has never been tested.

CWin32IOBufferedStreamWrapper::CWin32IOBufferedStreamWrapper(IStream *pStream)
: m_pStream(pStream), m_buffer(NULL), m_ulPosition(0), m_ulBufferSize(0)
{
}

CWin32IOBufferedStreamWrapper::~CWin32IOBufferedStreamWrapper()
{
	delete [] m_buffer;
	m_buffer = NULL;
}

size_t CWin32IOBufferedStreamWrapper::read(void *ptr, size_t size, size_t nmemb)
{
	if (m_ulPosition > m_ulBufferSize)
	{
		// The caller has seeked beyond the end of the file. We cannot read anything.
		errno = 0;
		return 0;
	}

	unsigned long ulBytesActuallyRead = 0;
	bool bError = false;

	unsigned long ulBytesToRead = nmemb;
	ulBytesToRead *= size;

	unsigned long ulBytesToCopy = m_ulBufferSize - m_ulPosition;

	if (ulBytesToCopy > ulBytesToRead)
	{
		ulBytesToCopy = ulBytesToRead;
	}

	if (ulBytesToCopy != 0)
	{
		memcpy(ptr, m_buffer + m_ulPosition, ulBytesToCopy);
		ulBytesActuallyRead += ulBytesToCopy;
		m_ulPosition = m_ulBufferSize;
	}

	ulBytesToRead -= ulBytesToCopy;

	if (ulBytesToRead != 0)
	{
		char *pOut = reinterpret_cast<char *>(ptr) + ulBytesToCopy;

		if (!reallocBuffer(m_ulBufferSize + ulBytesToRead))
		{
			bError = true;
		}
		else
		{
			readIntoBuffer(m_ulBufferSize, ulBytesToRead)
		}


	}
}

int CWin32IOBufferedStreamWrapper::seek(long offset, int whence)
{
	unsigned long ulNewPosition;

	bool bError = false;

	switch(whence)
	{
	default:
	case SEEK_SET:
		ulNewPosition = offset;
		break;
	case SEEK_CUR:
		ulNewPosition = m_ulPosition;
		ulNewPosition += offset;
		break;
	case SEEK_END:
		{
			STATSTG stats;
			if (S_OK != m_pStream->Stat(&stats, STATFLAG_NONAME)
			||	stats.cbSize.QuadPart > ULONG_MAX)
			{
				bError = true;
			}
			else
			{
				ulNewPosition = static_cast<unsigned long>(stats.cbSize.QuadPart);
				ulNewPosition += offset;
			}
		}
		break;
	}

	if (!bError)
	{
		if (m_ulBufferSize >= ulNewPosition)
		{
			// We've already been here so just note the new position.
			m_ulPosition = ulNewPosition;
		}
		else
		{
			// Read and buffer data up to the new position.
			unsigned long ulBytesToRead = ulNewPosition - m_ulBufferSize;

			unsigned long ulBytesRead = 0;

			// Neither reallocBuffer nor readIntoBuffer will change m_ulBufferSize or m_ulPosition.
			if (!reallocBuffer(ulNewPosition)
			||	!readIntoBuffer(m_ulBufferSize, ulBytesToRead, &ulBytesRead))
			{
				// If the allocation succeeded but the read failed we'll be left with extra memory
				// that we never really use. Oh well.
				bError = true;
			}
			else
			{
				// We were able to allocate and read the data so update the bufferSize and position.
				m_ulBufferSize = m_ulBufferSize + ulBytesRead;
				m_ulPosition = ulNewPosition; // Position may be past the end of the file/stream/buffer.
			}
		}
	}

	if (!bError)
	{
		return 0; // success
	}

	if (m_bThrowExceptions) { throw "seek failed"; }
	return -1; // error
}

long CWin32IOBufferedStreamWrapper::tell()
{
	if (m_ulPosition > LONG_MAX)
	{
		if (m_bThrowExceptions) { throw "tell failed"; }
		errno = EBADF;
		return -1L;
	}
	return m_ulPosition;
}

bool CWin32IOBufferedStreamWrapper::CanSeek()
{
	return true;
}
*/

CWin32IOStreamWrapper::CWin32IOStreamWrapper(IStream *pStream, bool bCanSeek)
: m_pStream(pStream), m_bCanSeek(bCanSeek)
{
}

CWin32IOStreamWrapper::~CWin32IOStreamWrapper()
{
}

size_t CWin32IOStreamWrapper::read(void *ptr, size_t size, size_t nmemb)
{
	ULONG ulBytesToRead = size;
	ulBytesToRead *= nmemb;

	ULONG ulBytesRead = 0;

	if (S_OK == m_pStream->Read(ptr, ulBytesToRead, &ulBytesRead))
	{
		errno = 0;
		return ulBytesRead;
	}

	errno = 1;
	if (m_bThrowExceptions) { throw "read failed"; }
	return 0;
}

int CWin32IOStreamWrapper::seek(long offset, int whence)
{
	if (m_bCanSeek)
	{
		LARGE_INTEGER liOffset;
		liOffset.QuadPart = offset;

		DWORD dwOrigin;

		switch(whence)
		{
		default:
		case SEEK_SET:	dwOrigin = STREAM_SEEK_SET;	break;
		case SEEK_CUR:	dwOrigin = STREAM_SEEK_CUR;	break;
		case SEEK_END:	dwOrigin = STREAM_SEEK_END;	break;
		}

		if (S_OK == m_pStream->Seek(liOffset, dwOrigin, NULL))
		{
			return 0; // success
		}
	}

	if (m_bThrowExceptions) { throw "seek failed"; }
	return -1; // error
}

long CWin32IOStreamWrapper::tell()
{
	if (m_bCanSeek)
	{
		LARGE_INTEGER liSeek;
		liSeek.QuadPart = 0;

		ULARGE_INTEGER uliPos;
		uliPos.QuadPart = 0;

		if (S_OK == m_pStream->Seek(liSeek, STREAM_SEEK_CUR, &uliPos))
		{
			return uliPos.LowPart; // success
		}
	}

	if (m_bThrowExceptions) { throw "tell failed"; }
	return -1; // error
}

bool CWin32IOStreamWrapper::CanSeek()
{
	return m_bCanSeek;
}




CWin32IOFileWrapper::CWin32IOFileWrapper(const char *szName, HANDLE hAbortEvent)
: m_hAbortEvent(hAbortEvent)
{
	m_hFile = ::CreateFile(szName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL);

#ifdef NJR_FILEWRAPDEBUG
	if (INVALID_HANDLE_VALUE == m_hFile
	||	NULL != strstr(szName, "_debug")
	||	NULL == strstr(szName, ".bmp"))
	{
		m_hDebugFile = INVALID_HANDLE_VALUE;
	}
	else
	{
		std::string strDebugName = szName;
		strDebugName += "_debug";

		m_hDebugFile = ::CreateFile(strDebugName.c_str(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

		if (INVALID_HANDLE_VALUE != m_hDebugFile)
		{
			BYTE zeros[1024];
			ZeroMemory(zeros, 1024);

			DWORD dwSize = ::GetFileSize(m_hFile, NULL);
			DWORD dwDontCare = 0;

			while (dwSize > 0)
			{
				DWORD dwToWrite = 1024;
				if (dwToWrite > dwSize)
				{
					dwToWrite = dwSize;
				}
				::WriteFile(m_hDebugFile, zeros, dwToWrite, &dwDontCare, NULL);
				dwSize -= dwToWrite;
			}

			::SetFilePointer(m_hDebugFile, 0, NULL, SEEK_SET);
		}
	}
#endif
}

CWin32IOFileWrapper::~CWin32IOFileWrapper()
{
	if (INVALID_HANDLE_VALUE != m_hFile)
	{
		CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
	}

#ifdef NJR_FILEWRAPDEBUG
	if (INVALID_HANDLE_VALUE != m_hDebugFile)
	{
		CloseHandle(m_hDebugFile);
		m_hDebugFile = INVALID_HANDLE_VALUE;
	}
#endif
}

bool CWin32IOFileWrapper::IsHandleValid()
{
	return (INVALID_HANDLE_VALUE != m_hFile);
}

size_t CWin32IOFileWrapper::read(void *ptr, size_t size, size_t nmemb)
{
	if (INVALID_HANDLE_VALUE != m_hFile)
	{
		DWORD dwBytesToRead = size * nmemb;
		DWORD dwBytesRead = 0;

		if (NULL != m_hAbortEvent && WAIT_OBJECT_0 == WaitForSingleObject(m_hAbortEvent, 0))
		{
			if (m_bThrowExceptions) { throw "abort event"; }
			return 0; // Indicate a read error.
		}

		if (::ReadFile(m_hFile, ptr, dwBytesToRead, &dwBytesRead, NULL))
		{
#ifdef NJR_FILEWRAPDEBUG
			if (INVALID_HANDLE_VALUE != m_hDebugFile)
			{
				DWORD dwDontCare = 0;
				::WriteFile(m_hDebugFile, ptr, dwBytesRead, &dwDontCare, NULL);
			}
#endif
			errno = 0;
			return dwBytesRead; // If dwBytesRead==0 then we are signalling the end of the input.
		}
	}

#ifdef NJR_FILEWRAPDEBUG
	MessageBeep(-1);
#endif

	errno = 1;
	if (m_bThrowExceptions) { throw "read failed"; }
	return 0; // Indicate a read error.
}

int CWin32IOFileWrapper::seek(long offset, int whence)
{
	if (INVALID_HANDLE_VALUE != m_hFile)
	{
		if (NULL != m_hAbortEvent && WAIT_OBJECT_0 == WaitForSingleObject(m_hAbortEvent, 0))
		{
			if (m_bThrowExceptions) { throw "abort event"; }
			return -1; // error
		}

		DWORD dwMoveMethod;

		switch(whence)
		{
		default:
		case SEEK_SET:	dwMoveMethod = FILE_BEGIN;		break;
		case SEEK_CUR:	dwMoveMethod = FILE_CURRENT;	break;
		case SEEK_END:	dwMoveMethod = FILE_END;		break;
		}

		DWORD dwResult = ::SetFilePointer(m_hFile, offset, NULL, dwMoveMethod);

		if (INVALID_SET_FILE_POINTER != dwResult || NO_ERROR == GetLastError())
		{
#ifdef NJR_FILEWRAPDEBUG
			if (INVALID_HANDLE_VALUE != m_hDebugFile)
			{
				::SetFilePointer(m_hDebugFile, offset, NULL, dwMoveMethod);
			}
#endif
			return 0; // success
		}
	}

#ifdef NJR_FILEWRAPDEBUG
	MessageBeep(-1);
#endif

	if (m_bThrowExceptions) { throw "seek failed"; }
	return -1; // error
}

long CWin32IOFileWrapper::tell()
{
	if (INVALID_HANDLE_VALUE != m_hFile)
	{
		if (NULL != m_hAbortEvent && WAIT_OBJECT_0 == WaitForSingleObject(m_hAbortEvent, 0))
		{
			if (m_bThrowExceptions) { throw "abort event"; }
			return -1; // Error.
		}

		DWORD dwResult = ::SetFilePointer(m_hFile, 0, NULL, FILE_CURRENT);

		if (INVALID_SET_FILE_POINTER != dwResult)
		{
			return dwResult; // Success
		}
	}

#ifdef NJR_FILEWRAPDEBUG
	MessageBeep(-1);
#endif

	if (m_bThrowExceptions) { throw "tell failed"; }
	return -1; // Error.
}

bool CWin32IOFileWrapper::CanSeek()
{
	return true;
}

void CWin32IOJpegSource::impl_init_source(j_decompress_ptr cinfo)
{
	CWin32IOJpegSource *pThis = reinterpret_cast<CWin32IOJpegSource *>(cinfo->src);

	pThis->m_bStartOfFile = true;
}

boolean CWin32IOJpegSource::impl_fill_input_buffer(j_decompress_ptr cinfo)
{
	CWin32IOJpegSource *pThis = reinterpret_cast<CWin32IOJpegSource *>(cinfo->src);

	size_t nbytes;

	try
	{
		nbytes = pThis->m_pFile->read(pThis->m_buffer, sizeof(JOCTET), WIJS_BUFFER_SIZE);
	}
	catch(...)
	{
		nbytes = 0;
	}

	if (nbytes <= 0)
	{
		if (pThis->m_bStartOfFile)
		{
			ERREXIT(cinfo, JERR_INPUT_EMPTY);
		}
		WARNMS(cinfo,JWRN_JPEG_EOF);
		pThis->m_buffer[0]=(JOCTET)0xFF;
		pThis->m_buffer[1]=(JOCTET)JPEG_EOI;
		nbytes=2;
	}
	else
	{
		_swab(reinterpret_cast<char *>(pThis->m_buffer), reinterpret_cast<char *>(pThis->m_buffer), nbytes);
	}

	pThis->next_input_byte = pThis->m_buffer;
	pThis->bytes_in_buffer = nbytes;
	pThis->m_bStartOfFile = false;

	return TRUE;
}

void CWin32IOJpegSource::impl_skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
	CWin32IOJpegSource *pThis = reinterpret_cast<CWin32IOJpegSource *>(cinfo->src);

	if (num_bytes > 0)
	{
		while (static_cast<size_t>(num_bytes) > pThis->bytes_in_buffer)
		{
			num_bytes -= static_cast<size_t>(pThis->bytes_in_buffer);
			impl_fill_input_buffer(cinfo);
		}
		pThis->next_input_byte += num_bytes;
		pThis->bytes_in_buffer -= num_bytes;
	}
}

void CWin32IOJpegSource::impl_term_source(j_decompress_ptr cinfo)
{
}

CWin32IOJpegSource::CWin32IOJpegSource(CWin32IOWrapper *pFile)
: m_pFile(pFile)
, m_bStartOfFile(true)
{
	m_buffer = new JOCTET[ WIJS_BUFFER_SIZE ];

	next_input_byte = NULL;
	bytes_in_buffer = 0;

	init_source = impl_init_source;
	fill_input_buffer = impl_fill_input_buffer;
	skip_input_data = impl_skip_input_data;
	resync_to_restart = jpeg_resync_to_restart;
	term_source = impl_term_source;
}

CWin32IOJpegSource::~CWin32IOJpegSource()
{
	delete [] m_buffer;
}
