// jp2raw.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "resource.h"
#include "LeoHelpers.h"
#include "jp2raw.h"
#include "JP2RawConfig.h"
#include "dcraw.h"
#include "Win32IOWrapper.h"

// {108DD0FC-E443-4d0d-A3F7-5473A5CE46B9}
static const GUID GUIDPlugin_jp2raw =
{ 0x108dd0fc, 0xe443, 0x4d0d, { 0xa3, 0xf7, 0x54, 0x73, 0xa5, 0xce, 0x46, 0xb9 } };


HMODULE g_hDllModule = NULL;
DWORD64 g_dw64OpusVersion = 0;
//UINT	g_uiMaxExtsLength = 0;
CRITICAL_SECTION g_lcmsCS;

HBITMAP loadBitmapViaXImage(const char *lpszName, LPVIEWERPLUGINFILEINFO lpVPFileInfo, LPSIZE lpszDesiredSize, HANDLE hAbortEvent);

// Note: We assume that Opus has initialized the common controls and COM for all
// threads which call our code. See gifanim plugin source for rant.

BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
	switch(ul_reason_for_call)
	{
	case(DLL_PROCESS_ATTACH):
		g_hDllModule = reinterpret_cast<HMODULE>(hModule);
		InitializeCriticalSection(&g_lcmsCS);
		break;
	case(DLL_PROCESS_DETACH):
		DeleteCriticalSection(&g_lcmsCS);
		break;
	default:
	case(DLL_THREAD_ATTACH):
	case(DLL_THREAD_DETACH):
		break;
	}

    return(TRUE);
}

DWORD64 getOpusVersionNumber()
{
//	EnterCriticalSection(&g_cs);

	if (0 == g_dw64OpusVersion)
	{
		TCHAR szModName[_MAX_PATH + 1];
		DWORD dwVerSize = 0;
		DWORD dwUnused;
		TCHAR *szParent       = NULL;
		TCHAR *szParentParent = NULL;
		TCHAR *szParentParentParent = NULL; // Needed for non-english versions.
		TCHAR *szOpusExePath  = NULL;

		if (NULL != g_hDllModule
		&&	GetModuleFileName(g_hDllModule, szModName, _MAX_PATH)
		&&	NULL != (szParent       = NLeoHelpers::GetParentPath(szModName))
		&&	NULL != (szParentParent = NLeoHelpers::GetParentPath(szParent))
		&&	NULL != (szOpusExePath  = NLeoHelpers::JoinPaths(szParentParent, _T("dopus.exe"))))
		{
			dwVerSize = GetFileVersionInfoSize(szOpusExePath, &dwUnused);

			if (0 == dwVerSize)
			{
				delete [] szOpusExePath;
				szOpusExePath = NULL;

				if (NULL != (szParentParentParent = NLeoHelpers::GetParentPath(szParentParent))
				&&	NULL != (szOpusExePath  = NLeoHelpers::JoinPaths(szParentParentParent, _T("dopus.exe"))))
				{
					dwVerSize = GetFileVersionInfoSize(szOpusExePath, &dwUnused);
				}
			}
		}

		if (0 != dwVerSize)
		{
			BYTE *pVerData = new BYTE[dwVerSize];

			if (GetFileVersionInfo(szOpusExePath, NULL, dwVerSize, pVerData))
			{
				LPVOID pBuffer;
				UINT bufLen;

				// Find out language of version resource
				if (VerQueryValue(pVerData, _T("\\VarFileInfo\\Translation"), &pBuffer, &bufLen)
				&&	4 == bufLen)
				{
					BYTE *pLang = reinterpret_cast<BYTE*>(pBuffer);

					DWORD dwHigh;
					DWORD dwLow;

					if (copyVersionNumber(&dwHigh, &dwLow, _T("ProductVersion"), pLang, pVerData))
					{
						g_dw64OpusVersion = dwHigh;
						g_dw64OpusVersion <<= 32;
						g_dw64OpusVersion |= dwLow;
					}
				}
			}

			delete [] pVerData;
		}

		delete [] szParent;
		delete [] szParentParent;
		delete [] szParentParentParent;
		delete [] szOpusExePath;
	}

	DWORD64 dw64Result = g_dw64OpusVersion;

//	LeaveCriticalSection(&g_cs);

	return(dw64Result);
}

bool copyVersionNumber(DWORD *pdwHigh, DWORD *pdwLow, TCHAR *szItemName, BYTE *pLang, const LPVOID pVerData)
{
	bool bResult = false;

	TCHAR szVerPath[128];
	_stprintf(szVerPath, _T("\\StringFileInfo\\%02x%02x%02x%02x\\%s"), pLang[1], pLang[0], pLang[3], pLang[2], szItemName);

	LPVOID pBuffer;
	UINT bufLen;
	int v0,v1,v2,v3;

	if (VerQueryValue(pVerData, szVerPath, &pBuffer, &bufLen)
	&&	4 == _stscanf(reinterpret_cast<TCHAR *>(pBuffer), _T("%d, %d, %d, %d"), &v0, &v1, &v2, &v3))
	{
		*pdwHigh = ((v0<<16)|v1);
		*pdwLow  = ((v2<<16)|v3);

		bResult = true;
	}

	return(bResult);
}

bool copyVersionString(TCHAR *szDestBuffer, UINT uiDestBufferSize, TCHAR *szItemName, BYTE *pLang, const LPVOID pVerData)
{
	bool bResult = false;

	if (NULL == szDestBuffer)
	{
		bResult = true;
	}
	else if (2 < uiDestBufferSize)
	{
		TCHAR szVerPath[128];
		_stprintf(szVerPath, _T("\\StringFileInfo\\%02x%02x%02x%02x\\%s"), pLang[1], pLang[0], pLang[3], pLang[2], szItemName);

		LPVOID pBuffer;
		UINT bufLen;

		if (VerQueryValue(pVerData, szVerPath, &pBuffer, &bufLen))
		{
			_tcsncpy(szDestBuffer, reinterpret_cast<TCHAR *>(pBuffer), uiDestBufferSize - 1);
			szDestBuffer[ uiDestBufferSize - 1 ] = _T('\0'); // strncpy/_tcsncpy sucks.

			bResult = true;
		}
	}

	return(bResult);
}

BOOL DVP_IdentifyA(LPVIEWERPLUGININFOA lpVPInfo)
{
	BOOL bResult = FALSE;

	DWORD64 dw64OpusVersion  = getOpusVersionNumber();

	if (NJP2RawConfig::CheckOpus(dw64OpusVersion, NJP2RawConfig::NTTA_RUN))
	{
		// DVPFIF_OverrideInternal is needed because some Raw camera images are in TIF format and Opus
		// recognises the file format but cannot handle the raw contents.

		lpVPInfo->dwFlags = DVPFIF_CanHandleStreams
						  | DVPFIF_CanConfigure
						  | DVPFIF_OverrideInternal;
					//	  | DVPFIF_InitialDisable;
					//	  | DVPFIF_NoFileInformation
					//	  | DVPFIF_ExtensionsOnly
					//	  | DVPFIF_ExtensionsOnlyIfSlow
					//	  | DVPFIF_ExtensionsOnlyForThumbnails


		std::vector< std::basic_string<TCHAR> > vecExts;
		NJP2RawConfig::RegGetRawExtensions(false, &vecExts);
		std::basic_string<TCHAR> strExts;

		vecExts.push_back( ".jp2" );
		vecExts.push_back( ".ppm" );
		vecExts.push_back( ".pnm" );
		vecExts.push_back( ".pgm" );
		vecExts.push_back( ".pbm" );
		vecExts.push_back( ".ras" );
		vecExts.push_back( ".tga" ); // Update Identify function if more TGA extensions are added, so they work with slow media.
		vecExts.push_back( ".tif" );
//		vecExts.push_back( ".tiff" );

		std::sort( vecExts.begin(), vecExts.end() );

		for(std::vector< std::basic_string<TCHAR> >::const_iterator pExt = vecExts.begin(); pExt != vecExts.end(); ++pExt)
		{
			if (!strExts.empty())
			{
				strExts += ';';
			}
			strExts += *pExt;
		}

		if (NULL != lpVPInfo->lpszHandleExts && 1 < lpVPInfo->cchHandleExtsMax)
		{
			_tcsncpy(lpVPInfo->lpszHandleExts, strExts.c_str(), lpVPInfo->cchHandleExtsMax - 1);
			lpVPInfo->lpszHandleExts[ lpVPInfo->cchHandleExtsMax - 1 ] = _T('\0'); // strncpy/_tcsncpy sucks.
		}

		lpVPInfo->dwlMinFileSize = 0;
		lpVPInfo->dwlMaxFileSize = 0;
		lpVPInfo->dwlMinPreviewFileSize = lpVPInfo->dwlMinFileSize;
		lpVPInfo->dwlMaxPreviewFileSize = lpVPInfo->dwlMaxFileSize;
		lpVPInfo->uiMajorFileType = DVPMajorType_Image;
		lpVPInfo->idPlugin = GUIDPlugin_jp2raw;

		// Get the remaining information from the version resource.

		TCHAR szModName[_MAX_PATH + 1];
		DWORD dwVerSize;
		DWORD dwUnused;

		if (NULL != g_hDllModule
		&&	GetModuleFileName(g_hDllModule, szModName, _MAX_PATH)
		&&	(0 != (dwVerSize = GetFileVersionInfoSize(szModName, &dwUnused))))
		{
			BYTE *pVerData = new BYTE[dwVerSize];

			if (GetFileVersionInfo(szModName, NULL, dwVerSize, pVerData))
			{
				LPVOID pBuffer;
				UINT bufLen;

				// Find out language of version resource
				if (VerQueryValue(pVerData, _T("\\VarFileInfo\\Translation"), &pBuffer, &bufLen)
				&&	4 == bufLen)
				{
					BYTE *pLang = reinterpret_cast<BYTE*>(pBuffer);

					if (copyVersionNumber(&lpVPInfo->dwVersionHigh,		&lpVPInfo->dwVersionLow,		_T("FileVersion"),		pLang, pVerData)
					&&	copyVersionString(lpVPInfo->lpszName,			lpVPInfo->cchNameMax,			_T("InternalName"),		pLang, pVerData)
					&&	copyVersionString(lpVPInfo->lpszDescription,	lpVPInfo->cchDescriptionMax,	_T("FileDescription"),	pLang, pVerData)
					&&	copyVersionString(lpVPInfo->lpszURL,			lpVPInfo->cchURLMax,			_T("CompanyName"   ),	pLang, pVerData)
					&&  copyVersionString(lpVPInfo->lpszCopyright,		lpVPInfo->cchCopyrightMax,		_T("LegalCopyright"),	pLang, pVerData))
					{
						bResult = TRUE;
					}
				}
			}
			delete [] pVerData;
		}
	}
	return(bResult);
}

bool pnm_getNextNumber(const unsigned char * & pBuf, const unsigned char * pEnd, int * piNumber)
{
	bool bResult = true;

	std::string strDigits;

	while(bResult)
	{
		if (pBuf >= pEnd)
		{
			bResult = false; // Failure: End of buffer.
		}
		else if ('#' == *pBuf || isspace(*pBuf))
		{
			if (!strDigits.empty())
			{
				*piNumber = atoi(strDigits.c_str());
				break;
			}

			if ('#' == *pBuf++)
			{
				// After a #, ignore everything else on the same line.
				while(pBuf < pEnd && '\n' != *pBuf && '\r' != *pBuf)
				{
					pBuf++;
				}
			}
		}
		else if (isdigit(*pBuf))
		{
			strDigits += *pBuf++;
		}
		else
		{
			bResult = false; // Failure: Unexpected character.
		}
	}

	return bResult;
}

bool jp2_getNextDWORD(const unsigned char * & pBuf, const unsigned char * pEnd, DWORD *pdw)
{
	if (4 > pEnd - pBuf)
	{
		return false;
	}

	DWORD a = static_cast<unsigned char>(*pBuf++);
	DWORD b = static_cast<unsigned char>(*pBuf++);
	DWORD c = static_cast<unsigned char>(*pBuf++);
	DWORD d = static_cast<unsigned char>(*pBuf++);

	*pdw = (a << 24) + (b << 16) + (c << 8) + d;

	return true;
}

bool jp2_getNextTypeChars(const unsigned char * & pBuf, const unsigned char * pEnd, char *typeBuffer)
{
	if (4 > pEnd - pBuf)
	{
		return false;
	}

	*typeBuffer++ = *pBuf++;
	*typeBuffer++ = *pBuf++;
	*typeBuffer++ = *pBuf++;
	*typeBuffer++ = *pBuf++;
	*typeBuffer = '\0';

	return true;
}

/*
bool jp2_getNext64Bit(const char * & pBuf, const char * pEnd, unsigned __int64 *pui64)
{
	if (8 > pEnd - pBuf)
	{
		return false;
	}

	unsigned __int64 a = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 b = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 c = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 d = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 e = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 f = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 g = static_cast<unsigned char>(*pBuf++);
	unsigned __int64 h = static_cast<unsigned char>(*pBuf++);

	*pui64 = (a << 56) + (b << 48) + (c << 40) + (d << 32) + (e << 24) + (f << 16) + (g << 8) + h;

	return true;
}
*/

bool jp2_getNextBox(const unsigned char * & pBuf, const unsigned char * pEnd, char * szBoxType, const unsigned char * * ppBoxData, DWORD * pdwBoxDataLength, bool *pbTruncated)
{
	bool bResult = false;

	DWORD dwBoxLength;
	const unsigned char *pOriginal = pBuf;

	if (jp2_getNextDWORD(pBuf, pEnd, &dwBoxLength)
	&&	jp2_getNextTypeChars(pBuf, pEnd, szBoxType)
	&&	8 <= dwBoxLength)	// JasPer cannot cope with dwBoxLength==1 (implies 64-bit length needs reading) so I won't bother either.
	{
		// If the box went off the end of our buffer, truncate whatever we have. The caller should check lengths.
		// This better than returning an error because a superbox may be huge but still contain the sub-boxes we need within the buffer.
		if (dwBoxLength > static_cast<DWORD>((pEnd - pOriginal)))
		{
			dwBoxLength = static_cast<DWORD>(pEnd - pOriginal);
			if (NULL != pbTruncated) { *pbTruncated = true; }
		}
		else
		{
			if (NULL != pbTruncated) { *pbTruncated = false; }
		}

		*ppBoxData = pBuf;
		pBuf = pOriginal + dwBoxLength;
		*pdwBoxDataLength = pBuf - *ppBoxData;
		bResult = true;
	}

	return bResult;
}

bool read32FromBuffer(DWORD *pdwResult, const unsigned char *pBufStart, const size_t bytesInBuffer, const size_t offset, const bool bSwap)
{
	if ((offset < 0) || (offset + sizeof(DWORD) > bytesInBuffer))
	{
		return false;
	}

	*pdwResult = *reinterpret_cast<const DWORD *>(pBufStart + offset);

	if (bSwap)
	{
		*pdwResult = ((((*pdwResult)&0x000000FF)<<24)
					 +(((*pdwResult)&0x0000FF00)<<8 )
					 +(((*pdwResult)&0x00FF0000)>>8 )
					 +(((*pdwResult)&0xFF000000)>>24));
	}

	return true;
}

bool read16FromBuffer(WORD *pwResult, const unsigned char *pBufStart, const size_t bytesInBuffer, const size_t offset, const bool bSwap)
{
	if ((offset < 0) || (offset + sizeof(WORD) > bytesInBuffer))
	{
		return false;
	}

	*pwResult = *reinterpret_cast<const WORD *>(pBufStart + offset);

	if (bSwap)
	{
		*pwResult = ((((*pwResult)&0x00FF)<<8)
					+(((*pwResult)&0xFF00)>>8 ));
	}

	return true;
}

BOOL identifyFileCommon(CWin32IOWrapper *pFileWrapper, const char *szName, bool bPhysicalFile, bool bSlowMedia, LPVIEWERPLUGINFILEINFOA lpVPFileInfo, HANDLE hAbortEvent)
{
	BOOL bResult = false;

	lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_UNKNOWN;

	pFileWrapper->SetThrowExceptions(true);

	DWORD dwConfigFlags = NJP2RawConfig::RegGetFlags(false);

	// Check for RAW extension and match first as RAW files may use the TIFF file structure.

	if (!bResult && !(dwConfigFlags & NJP2RawConfig::NJRF_DISABLE_DCRAW))
	{
		bool bRawFileExtMatch = false;

		// The Raw identify function is prone to false positives so we require a file extension match first.
		// This isn't a big deal since people can add to the list of extensions via the configuration dialog.
		if (NULL != szName)
		{
			size_t nameLen = strlen(szName);
			std::basic_string<TCHAR>::size_type extLen = 0;

			std::vector< std::basic_string<TCHAR> > vecExts;
			NJP2RawConfig::RegGetRawExtensions(false, &vecExts);

			for (std::vector< std::basic_string<TCHAR> >::const_iterator pExt = vecExts.begin(); pExt != vecExts.end(); ++pExt)
			{
				extLen = pExt->length();
				if (extLen <= nameLen
				&&	0 == _stricmp(pExt->c_str(), (szName + nameLen) - extLen))
				{
					bRawFileExtMatch = true;
					break;
				}
			}
		}

		if (bRawFileExtMatch)
		{
			if (!pFileWrapper->CanSeek() || bSlowMedia)
			{
				// A file extension match is good enough for media that is slow or cannot seek.
				// DCRaw won't work on non-seekable streams and the identify function is very slow if seeking is slow since it
				// has to move around the file a lot.
				lpVPFileInfo->szImageSize.cx = 0;
				lpVPFileInfo->szImageSize.cy = 0;
				lpVPFileInfo->iNumBits = 0;

				if (NULL != lpVPFileInfo->lpszInfo)
				{
					NLeoHelpers::StringCopy(lpVPFileInfo->lpszInfo, "Raw Image", lpVPFileInfo->cchInfoMax);
				}

				bResult = true;
			}
			else
			{
				try
				{
					pFileWrapper->seek(0, SEEK_SET); // Rewind buffer reading as a result of other checks.

					std::string strMake;
					std::string strModel;
					int iWidth, iHeight;

					// NULL output name and settings means we just identify the file (quick) and don't decode the image inside it (slow).
					DCRaw *pdcr = new DCRaw(&g_lcmsCS, hAbortEvent); // Do not create this on the stack. It is huge.
					if (pdcr->DoFile(NULL, pFileWrapper, bPhysicalFile ? szName : NULL, NULL, &strMake, &strModel, &iWidth, &iHeight))
					{
						lpVPFileInfo->szImageSize.cx = iWidth;
						lpVPFileInfo->szImageSize.cy = iHeight;
						lpVPFileInfo->iNumBits = 0;

						if (NULL != lpVPFileInfo->lpszInfo)
						{
							if (strMake.empty() && strModel.empty())
							{
								NLeoHelpers::StringFormat(lpVPFileInfo->lpszInfo, lpVPFileInfo->cchInfoMax, "%ld x %ld Raw Image",
															lpVPFileInfo->szImageSize.cx, lpVPFileInfo->szImageSize.cy);
							}
							else
							{
								// If the make is repeated in the model, only print it once.
								if (0 == _stricmp(strMake.c_str(), strModel.substr(0, strMake.length()).c_str())
								&&	(strMake.length() == strModel.length() || isspace(strModel[strMake.length()])))
								{
									strMake.clear();
								}

								NLeoHelpers::StringFormat(lpVPFileInfo->lpszInfo, lpVPFileInfo->cchInfoMax, "%ld x %ld Raw Image from %s%s%s",
															lpVPFileInfo->szImageSize.cx, lpVPFileInfo->szImageSize.cy,
															strMake.c_str(), !(strMake.empty()||strModel.empty()) ? " " : "", strModel.c_str());
							}
						}

						bResult = true;
					}
					delete pdcr;
				}
				catch(...)
				{
					bResult = false;
				}
			}

			if (bResult)
			{
				lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_DCRAW;
			}
		}
	}

	DWORD dwDisableAllButRaw = (NJP2RawConfig::NJRF_DISABLE_TGA
		                      | NJP2RawConfig::NJRF_DISABLE_JP2
							  | NJP2RawConfig::NJRF_DISABLE_PNM
							  | NJP2RawConfig::NJRF_DISABLE_RAS);
						//	  | NJP2RawConfig::NJRF_DISABLE_TIF);

	if (!bResult && (dwConfigFlags & dwDisableAllButRaw) != dwDisableAllButRaw)
	{
		// Read the start of the file and inspect it to see if it's a JP2, TGA, TIFF or PNM image.
		try
		{
			char unsigned buffer[1024]; // JP2 headers can be quite large, particulary if a stupid program puts a load of XML before the IHDR. Sigh.
			const unsigned char *endOfBuffer = buffer + sizeof(buffer);
			ZeroMemory(buffer, sizeof(buffer));

			if (pFileWrapper->CanSeek() && !bSlowMedia)
			{
				pFileWrapper->seek(0, SEEK_SET); // Rewind buffer reading as a result of other checks.
			}

			size_t bytesInBuffer = pFileWrapper->read(buffer, 1, sizeof(buffer));

			if (0 < bytesInBuffer)
			{
				const char *szType = "";
				int iWidth, iHeight, iNumBits;
				int iPages = 0;
				bool bGotDims = false;
				bool bGotBits = false;
				bool bGotPages = false;
				bool bGotPagesExact = false;

				// JP2
				if (!bResult && !(dwConfigFlags & NJP2RawConfig::NJRF_DISABLE_JP2))
				{
					szType = ""; iWidth = iHeight = iNumBits = 0;
					const unsigned char *pb = buffer;

					char szBoxType[5];
					const unsigned char *pBoxData;
					DWORD dwBoxDataLength;
					if (jp2_getNextBox(pb, endOfBuffer, szBoxType, &pBoxData, &dwBoxDataLength, NULL)
					&&	0 == strcmp("jP  ", szBoxType)
					&&	dwBoxDataLength == 4
					&&	pBoxData[0] == 0x0D && pBoxData[1] == 0x0A && pBoxData[2] == 0x87 && pBoxData[3] == 0x0A
					&&	jp2_getNextBox(pb, endOfBuffer, szBoxType, &pBoxData, &dwBoxDataLength, NULL)
					&&	0 == strcmp("ftyp", szBoxType)) // Jasper doesn't check the contents of the ftyp box so neither will I.
					{
						szType = "JPEG 2000";
						bResult = true;
						lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_JP2;

						bool bTruncated;

						// See if we can find the dimensions and bit depth.
						// Find the jp2h super-box.
						while(jp2_getNextBox(pb, endOfBuffer, szBoxType, &pBoxData, &dwBoxDataLength, NULL))
						{
							if (0 == strcmp("jp2h", szBoxType))
							{
								const unsigned char *pCurrentSubBox = pBoxData;
								const unsigned char *pEndOfSubBoxes = pCurrentSubBox + dwBoxDataLength;

								bool bGotIhdrBits = false;

								while(!bGotIhdrBits && jp2_getNextBox(pCurrentSubBox, pEndOfSubBoxes, szBoxType, &pBoxData, &dwBoxDataLength, &bTruncated))
								{
									if (0 == strcmp("ihdr", szBoxType) && sizeof(JP2IhdrBox) == dwBoxDataLength && !bTruncated)
									{
										JP2IhdrBox ihdr(pBoxData);

										iWidth = ihdr.width;
										iHeight = ihdr.height;
										bGotDims = true;

										if (0 < ihdr.bpc)
										{
											iNumBits = ihdr.numcmpts * ((ihdr.bpc & 0x7F) + 1); // Strip high bit which represents whether components are signed.
											bGotBits = true;
											bGotIhdrBits = true;
										}
									}
									//else if (0 == strcmp("bpcc", szBoxType && !bTruncated)
									//{
									//	// Data of bcpp box is one byte per component.
									//	// Really not sure if this is 7-bit -1 as above or what, though, and got no test images...
									//	// ...so just not going to report the depth of this kind of image for now.
									//}
								}

								break;
							}
						}
					}
				}

				// PNM
				if (!bResult && !(dwConfigFlags & NJP2RawConfig::NJRF_DISABLE_PNM) && bytesInBuffer > 7)
				{
					szType = ""; iWidth = iHeight = iNumBits = 0;
					const unsigned char *pb = buffer + 3;
					int iMaxColor;

					if (buffer[0] == 'P' && isdigit(buffer[1]) && isspace(buffer[2])
					&&	pnm_getNextNumber(pb, endOfBuffer, &iWidth)
					&&	pnm_getNextNumber(pb, endOfBuffer, &iHeight))
					{
						switch(buffer[1])
						{
						case('1'):
						case('4'):
							iNumBits = 1;
							szType = "PBM";
							bGotDims = true;
							bGotBits = true;
							bResult = true;
							break;
						case('2'):
						case('5'):
						case('3'):
						case('6'):
							if (pnm_getNextNumber(pb, endOfBuffer, &iMaxColor)
							&&	0 < iMaxColor && iMaxColor < 65536)
							{
								if      (iMaxColor >= 32768) { iNumBits = 16; }
							    else if (iMaxColor >= 16385) { iNumBits = 15; }
                                else if (iMaxColor >=  8192) { iNumBits = 14; }
                                else if (iMaxColor >=  4096) { iNumBits = 13; }
                                else if (iMaxColor >=  2048) { iNumBits = 12; }
							    else if (iMaxColor >=  1024) { iNumBits = 11; }
								else if (iMaxColor >=   512) { iNumBits = 10; }
								else if (iMaxColor >=   256) { iNumBits =  8; }
							    else if (iMaxColor >=   128) { iNumBits =  8; }
								else if (iMaxColor >=    64) { iNumBits =  7; }
								else if (iMaxColor >=    32) { iNumBits =  6; }
								else if (iMaxColor >=    16) { iNumBits =  5; }
								else if (iMaxColor >=     8) { iNumBits =  4; }
								else if (iMaxColor >=     4) { iNumBits =  3; }
								else if (iMaxColor >=     2) { iNumBits =  2; }
								else if (iMaxColor >=     1) { iNumBits =  1; }
								else                         { iNumBits =  0; }

								if(buffer[1] == '2' || buffer[1] == '5')
								{
									szType = "PGM";
								}
								else
								{
									szType = "PPM";
									iNumBits *= 3;
								}
								bGotDims = true;
								bGotBits = true;
								bResult = true;
							}
							break;
						default:
							// Not a recognised PNM type.
							break;
						}
					}

					if (bResult)
					{
						lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_PNM;
					}
				}

				// RAS
				if (!bResult && !(dwConfigFlags & NJP2RawConfig::NJRF_DISABLE_RAS) && bytesInBuffer >= sizeof(RasterfileHeader))
				{
					RasterfileHeader rasHead(buffer);

					if (rasHead.ras_magic == 0x59a66a95)
					{
						iWidth = rasHead.ras_width;
						iHeight = rasHead.ras_height;
						iNumBits = rasHead.ras_depth;
						szType = "Rasterfile";
						bGotDims = true;
						bGotBits = true;
						bResult = true;
						lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_RAS;
					}
				}
#if 0 // REMINDER: Search for NJRF_DISABLE_TIF where commented-out.
				// TIFF
				if (!bResult && !(dwConfigFlags & NJP2RawConfig::NJRF_DISABLE_TIF) && bytesInBuffer >= 20)
				{
					bool bSwap = false;
					bool bGotWidth = false;
					bool bGotHeight = false;

					if (buffer[0] == 'M' && buffer[1] == 'M' && buffer[2] == '\0' && buffer[3] == 42)
					{
						bSwap = true;
					}

					if (bSwap || (buffer[0] == 'I' && buffer[1] == 'I' && buffer[2] == 42 && buffer[3] == '\0'))
					{
						szType = "TIFF MF!!!";
						bResult = true;
						lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_TIF;

						DWORD dwDirOffset;
						DWORD dwDirEntryOffset;
						WORD wDirEntryCount;
						WORD wDirEntryTag;
						WORD wDirEntryType;
						DWORD dwDirEntryValueCount;
						DWORD dwDirEntryValue;
						WORD wDirEntryValue;
						
						if (read32FromBuffer(&dwDirOffset, buffer, bytesInBuffer, 4, bSwap)
						&&	read16FromBuffer(&wDirEntryCount, buffer, bytesInBuffer, dwDirOffset, bSwap)
						&&	wDirEntryCount > 0)
						{
							const WORD wTifType_Byte  = 1;
							const WORD wTifType_Short = 3;
							const WORD wTifType_Long  = 4;

							for (WORD wDirEntryIndex = 0; wDirEntryIndex < wDirEntryCount; ++wDirEntryIndex)
							{
								dwDirEntryOffset = dwDirOffset + 2 + wDirEntryIndex * 12;

								if (!read16FromBuffer(&wDirEntryTag, buffer, bytesInBuffer, dwDirEntryOffset, bSwap)
								||	!read16FromBuffer(&wDirEntryType, buffer, bytesInBuffer, dwDirEntryOffset + 2, bSwap)
								||	!read32FromBuffer(&dwDirEntryValueCount, buffer, bytesInBuffer, dwDirEntryOffset + 4, bSwap))
								{
									break;
								}

								if (wDirEntryTag == 0x100)// && wDirEntryCount == 1)
								{
									if (wDirEntryType == wTifType_Long && read32FromBuffer(&dwDirEntryValue, buffer, bytesInBuffer, dwDirEntryOffset + 8, bSwap))
									{
										iWidth = dwDirEntryValue;
										bGotWidth = true;
									}
									else if (wDirEntryType == wTifType_Short && read16FromBuffer(&wDirEntryValue, buffer, bytesInBuffer, dwDirEntryOffset + 8, bSwap))
									{
										iWidth = wDirEntryValue;
										bGotWidth = true;
									}
								}
								else if (wDirEntryTag == 0x101)// && wDirEntryCount == 1)
								{
									if (wDirEntryType == wTifType_Long && read32FromBuffer(&dwDirEntryValue, buffer, bytesInBuffer, dwDirEntryOffset + 8, bSwap))
									{
										iHeight = dwDirEntryValue;
										bGotHeight = true;
									}
									else if (wDirEntryType == wTifType_Short && read16FromBuffer(&wDirEntryValue, buffer, bytesInBuffer, dwDirEntryOffset + 8, bSwap))
									{
										iHeight = wDirEntryValue;
										bGotHeight = true;
									}
								}
							}

							if (bGotWidth && bGotHeight)
							{
								bGotDims = true;
							}

							iPages = 1;
							bGotPages = true;

							while (read32FromBuffer(&dwDirOffset, buffer, bytesInBuffer, dwDirOffset + 2 + wDirEntryCount * 12, bSwap))
							{
								if (dwDirOffset == 0)
								{
									bGotPagesExact = true;
									break;
								}

								if (!read16FromBuffer(&wDirEntryCount, buffer, bytesInBuffer, dwDirOffset, bSwap))
								{
									break;
								}
							}
						}

						// iNumBits = ...
						bGotBits = false;
					}
				}
#endif
				// TGA
				if (!bResult && !(dwConfigFlags & NJP2RawConfig::NJRF_DISABLE_TGA) && bytesInBuffer >= sizeof(TGAFileHeader))
				{
					szType = ""; iWidth = iHeight = iNumBits = 0;

					// Newer TGA files have a signature at the end of them, but not all and it's expensive to look at the end of the file
					// if we're on slow media or unable to seek.

					bool bLooksLikeTGA = false;

					// If the filename has a .tga extension, let's assume it's a TGA file and not waste any more time.
					if (NULL != szName)
					{
						size_t nameLen = strlen(szName);
						if (nameLen > 4)
						{
							const char *szLast4 = (szName + nameLen) - 4;
							if (0 == _stricmp(".tga", szLast4))
							{
								bLooksLikeTGA = true;
							}
						}
					}

					// If the filename doesn't have a .tga extension, check for the signature but only if we can seek and aren't on slow media.
					if (!bLooksLikeTGA && pFileWrapper->CanSeek() && !bSlowMedia)
					{
						try
						{
							pFileWrapper->seek(-18, SEEK_END);

							char tgaSigBuffer[17];
							if (16 == pFileWrapper->read(tgaSigBuffer, 1, 16))
							{
								tgaSigBuffer[16] = '\0';

								if (0 == strcmp("TRUEVISION-XFILE", tgaSigBuffer))
								{
									bLooksLikeTGA = true;
								}
							}
						}
						catch(...)
						{
							bLooksLikeTGA = false;
						}
					}

					if (bLooksLikeTGA)
					{
						const TGAFileHeader *pTgaHeader = reinterpret_cast<TGAFileHeader *>(buffer);

						iWidth   = pTgaHeader->wImageWidth;
						iHeight  = pTgaHeader->wImageHeight;
						iNumBits = pTgaHeader->bPixelDepth;
						szType = "TGA";
						bGotDims = true;
						bGotBits = true;
						bResult = true;
						lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ] = PDIT_TGA;
					}
				}

				if (bResult)
				{
					lpVPFileInfo->szImageSize.cx = iWidth;
					lpVPFileInfo->szImageSize.cy = iHeight;
					lpVPFileInfo->iNumBits = iNumBits;

					if (NULL != lpVPFileInfo->lpszInfo)
					{
						int nullIdx;

						if (bGotDims && bGotBits)
						{
							nullIdx = NLeoHelpers::StringFormat(lpVPFileInfo->lpszInfo, lpVPFileInfo->cchInfoMax, "%ld x %ld x %d %s Image",
																lpVPFileInfo->szImageSize.cx, lpVPFileInfo->szImageSize.cy, lpVPFileInfo->iNumBits, szType);
						}
						else if (bGotDims)
						{
							nullIdx = NLeoHelpers::StringFormat(lpVPFileInfo->lpszInfo, lpVPFileInfo->cchInfoMax, "%ld x %ld %s Image",
																lpVPFileInfo->szImageSize.cx, lpVPFileInfo->szImageSize.cy, szType);
						}
						else
						{
							nullIdx = NLeoHelpers::StringFormat(lpVPFileInfo->lpszInfo, lpVPFileInfo->cchInfoMax, "%s Image", szType);
						}

						if (bGotPages && iPages > 1)
						{
							NLeoHelpers::StringFormat(lpVPFileInfo->lpszInfo + nullIdx, lpVPFileInfo->cchInfoMax - nullIdx, " (%d%s pages)", iPages, bGotPagesExact ? "" : "+");
						}
					}
				}
			}
		}
		catch(...)
		{
			bResult = false;
		}
	}

	if (bResult)
	{
		lpVPFileInfo->dwFlags = DVPFIF_CanReturnBitmap
							  | DVPFIF_CanReturnThumbnail;

		lpVPFileInfo->wMajorType = DVPMajorType_Image;
		lpVPFileInfo->wMinorType = 0;
	}

	return bResult;
}

BOOL DVP_IdentifyFileA(HWND hWnd,LPTSTR lpszName,LPVIEWERPLUGINFILEINFOA lpVPFileInfo,HANDLE hAbortEvent)
{
	CWin32IOFileWrapper fileWrapper(lpszName, hAbortEvent);

	return identifyFileCommon(&fileWrapper, lpszName, true, false, lpVPFileInfo, hAbortEvent);
}
BOOL DVP_IdentifyFileStreamA(HWND hWnd,LPSTREAM lpStream,LPTSTR lpszName,LPVIEWERPLUGINFILEINFOA lpVPFileInfo,DWORD dwStreamFlags)
{
	CWin32IOStreamWrapper fileWrapper(lpStream, dwStreamFlags & DVPSF_NoRandomSeek ? true : false);

	// Set the slow flag for anything that's coming as a stream. Opus won't flag Zip streams, for example, but they are quite slow
	// to seek around in. (Not as slow as an FTP site, but slow enough that seeking to the end of a file isn't sensible.)
	// dwStreamFlags & DVPSF_Slow ? true : false
	return identifyFileCommon(&fileWrapper, lpszName, false, true, lpVPFileInfo, NULL);
}
HWND DVP_Configure(HWND hWndParent,HWND hWndNotify,DWORD dwNotifyData)
{
	// Structure will be copied, so creating it on the stack is fine, and if the create fails we don't leak anything.
	NJP2RawConfig::NJR_ConfigData cd;
	cd.hInstance		= g_hDllModule;
	cd.pLcmsCS			= &g_lcmsCS;
	cd.hWndNotify		= hWndNotify;
	cd.dwNotifyData		= dwNotifyData;
	cd.dw64OpusVersion	= getOpusVersionNumber();

	return(CreateDialogParam(g_hDllModule, MAKEINTRESOURCE(IDD_CONFIG), hWndParent, NJP2RawConfig::configDlgProc, reinterpret_cast<LPARAM>(&cd)));
}

HBITMAP DVP_LoadBitmapStreamA(HWND hWnd, LPSTREAM lpStream, LPTSTR lpszName, LPVIEWERPLUGINFILEINFO lpVPFileInfo, LPSIZE lpszDesiredSize, DWORD dwStreamFlags)
{
	// Even for decoders which can or do take the Win32IOWrapper, it's often always going to be quicker to extract the entire stream to a temp file
	// and then work on that, particularly with things like the Raw decoder which seek around a lot.

	HBITMAP hbmResult = NULL;

	// Get temporary filename

	TCHAR szTempPath[MAX_PATH];

	if (0 != GetTempPath(MAX_PATH, szTempPath))
	{
		std::basic_string< TCHAR > strTempFile;

		// Open temporary file for output
		HANDLE hFile = NLeoHelpers::OpenTempFileNamePreserveExtension(szTempPath, TEXT("dop"), lpszName, &strTempFile);

		if (INVALID_HANDLE_VALUE != hFile)
		{
			// Copy data to temporary file

			bool bWriteError = false;
			DWORD dwSize;
			BYTE bBuf[8192];

			while (S_OK == lpStream->Read(bBuf, sizeof(bBuf), &dwSize))
			{
				if (0 < dwSize)
				{
					DWORD dwTemp = 0;

					if (0 == WriteFile(hFile,bBuf,dwSize,&dwTemp,0))
					{
						bWriteError = true;
						break;
					}
				}
			}

			// Close temporary file
			CloseHandle(hFile);

			if (!bWriteError)
			{
				// Load new file
				hbmResult = DVP_LoadBitmapA(hWnd, const_cast< LPTSTR >( strTempFile.c_str() ), lpVPFileInfo, lpszDesiredSize, NULL);
			}

			DeleteFile( strTempFile.c_str() );
		}
	}

	return(hbmResult);
}

HBITMAP DVP_LoadBitmapA(HWND hWnd, LPTSTR lpszName, LPVIEWERPLUGINFILEINFO lpVPFileInfo, LPSIZE lpszDesiredSize, HANDLE hAbortEvent)
{
	HBITMAP hbmResult = NULL;

	bool bThumbnail = (NULL != lpszDesiredSize && 0 != lpszDesiredSize->cx && 0 != lpszDesiredSize->cy);

	if (PDIT_DCRAW != lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ])
	{
		hbmResult = loadBitmapViaXImage(lpszName, lpVPFileInfo, lpszDesiredSize, hAbortEvent);
	}
	else
	{
		// TODO: Remove the need to load all the configuration data every time. Maybe store it in a global, CS-protected map.
		NJP2RawConfig::NJR_ProfileMap mapProfiles;
		NJP2RawConfig::RegGetRawProfiles(false, mapProfiles);

		std::string strMake;
		std::string strModel;

		// This is a bit lame, but we need to know the make/model of the file and haven't stored it anywhere, so do another identify pass.
		// TODO: Remove the need to do this. Maybe point to an item in a global, CS-protected map with a unique name (in case it's removed).
		{
			CWin32IOFileWrapper inputFile(lpszName, hAbortEvent);

			// NULL settings and output name means we just identify the file (quick) and don't decode the image inside it (slow).
			DCRaw *pdcr = new DCRaw(&g_lcmsCS, hAbortEvent); // Do not create this on the stack. It is huge.
			if (!pdcr->DoFile(NULL, &inputFile, lpszName, NULL, &strMake, &strModel, NULL, NULL))
			{
				strMake.clear();
				strModel.clear();
			}
			delete pdcr;
		}

		NJP2RawConfig::NJR_RawProfile &rawProfile = mapProfiles.GetCameraProfileFromMakeModel(strMake.c_str(), strModel.c_str());
		NJP2RawConfig::NJR_RawSettings &rawSettings = bThumbnail ? rawProfile.rsThumbs : rawProfile.rsViewers;

		//rawSettings.dwFlags -= NJP2RawConfig::RAWF_TRY_PREVIEW;
		//rawSettings.dwFlags += NJP2RawConfig::RAWF_TRY_FULL;

		TCHAR szTempPath[MAX_PATH];

		if (0 != GetTempPath(MAX_PATH, szTempPath))
		{
			std::basic_string< TCHAR > strTempFileDecoded;

			// Open temporary file for output
			HANDLE hFile = NLeoHelpers::OpenTempFileNamePreserveExtension(szTempPath, TEXT("dop"), bThumbnail ? TEXT("output.thumb") : TEXT("output.ppm"), &strTempFileDecoded);

			if (INVALID_HANDLE_VALUE != hFile)
			{
				CloseHandle(hFile);
				hFile = INVALID_HANDLE_VALUE;

				// Scope file object
				{
					CWin32IOFileWrapper inputFile(lpszName, hAbortEvent);

					DCRaw *pdcr = new DCRaw(&g_lcmsCS, hAbortEvent); // Do not create this on the stack. It is huge.
					if (pdcr->DoFile(&rawSettings, &inputFile, lpszName, strTempFileDecoded.c_str(), NULL, NULL, NULL, NULL))
					{
						hbmResult = loadBitmapViaXImage(strTempFileDecoded.c_str(), lpVPFileInfo, lpszDesiredSize, hAbortEvent);
					}
					delete pdcr;
				}

				DeleteFile(strTempFileDecoded.c_str());
			}
		}
	}

	return(hbmResult);
}

HBITMAP loadBitmapViaXImage(const char *lpszName, LPVIEWERPLUGINFILEINFO lpVPFileInfo, LPSIZE lpszDesiredSize, HANDLE hAbortEvent)
{
	HBITMAP hbmResult = NULL;

	CxImage xi;

	DWORD dwImageType;
	
	switch(lpVPFileInfo->dwPrivateData[ PRIVATEDATA_IMAGETYPE_DWORD ])
	{
	case(PDIT_DCRAW):	dwImageType = CXIMAGE_FORMAT_UNKNOWN;	break;	// Will probably have been converted to PPM or JPG format. Let CxImage work it out.
	case(PDIT_JP2):		dwImageType = CXIMAGE_FORMAT_JP2;		break;
	case(PDIT_TGA):		dwImageType = CXIMAGE_FORMAT_TGA;		break;
	case(PDIT_PNM):		dwImageType = CXIMAGE_FORMAT_PNM;		break;
	case(PDIT_RAS):		dwImageType = CXIMAGE_FORMAT_RAS;		break;
	default:			dwImageType = CXIMAGE_FORMAT_UNKNOWN;	break;	// Should never happen, but let CxImage try to work out the image if it does.
	}

	if (xi.Load(lpszName, dwImageType))
	{
		if (xi.AlphaIsValid())
		{
			lpVPFileInfo->dwFlags |= DVPFIF_HasAlphaChannel;

			hbmResult = NLeoHelpers::MakeBitmapWithAlpha(xi, false);
		}
		else
		{
			hbmResult = xi.MakeBitmap(NULL);
		}
	}

	return(hbmResult);
}
